# ------------------------------------------------------------------------------
# Copyright (c) Microsoft
# Licensed under the MIT License.
# Written by Zhipeng Zhang (zhangzhipeng2017@ia.ac.cn)
# ------------------------------------------------------------------------------

import torch
import torch.nn as nn
import numpy as np

class AdaFree_(nn.Module):
    def __init__(self):
        super(AdaFree_, self).__init__()
        self.features = None
        self.connect_model0 = None
        self.connect_model1 = None
        self.align_head = None
        self.mask_model = None
        self.connect_model2 = None
        self.zf = None  # for online tracking
        # self.bbox_tower = None
        self.neck = None
        self.search_size = 255
        self.score_size = 25
        self.batch = 32 if self.training else 1
        self.lambda_u = 0.1
        self.lambda_s = 0.2


    def feature_extractor(self, x, online=False):
        return self.features(x, online=online)

    def tensorrt_init(self, trt_net, corr=None):
        self.t_backbone255, self.s_backbone_siam255, self.s_backbone_siam287, self.s_backbone_online, self.t_neck255, \
        self.s_neck255, self.s_neck287, self.multiDiCorr255, self.multiDiCorr287, self.boxtower255, self.boxtower287 = trt_net

        if corr:
            self.multiDiCorr255, self.multiDiCorr287 = corr

    def template(self, z, template_mask):
        _, _, _, self.zf = self.t_backbone255(z, torch.Tensor([]).cuda())
        self.zf_ori = self.t_neck255(self.zf)
        self.zf = self.zf_ori[:, :, 4:-4, 4:-4].contiguous()
        self.template_mask = template_mask.float()  # [1, H, W]
        self.MA_kernel = self.zf.detach()
        self.zf_update = None


    def template_pytorch(self, z, template_mask):
        _, _, _,  self.zf = self.features(z)
        self.zf_ori = self.neck(self.zf)
        self.zf = self.zf_ori[:, :, 4:-4, 4:-4]
        self.template_mask = template_mask.float()

    def update_roi_template(self, target_pos, target_sz, score):
        """
        :param target_pos:  pos in search (not the original)
        :param target_sz:  size in target size
        :param score:
        :return:
        """

        lambda_u = self.lambda_u * float(score)
        lambda_s = self.lambda_s
        N, C, H, W = self.search_size
        stride = 8
        assert N == 1, "not supported"
        l = W // 2
        x = range(-l, l + 1)
        y = range(-l, l + 1)

        hc_z = (target_sz[1] + 0.3 * sum(target_sz)) / stride
        wc_z = (target_sz[0] + 0.3 * sum(target_sz)) / stride
        grid_x = np.linspace(- wc_z / 2, wc_z / 2, 17)
        grid_y = np.linspace(- hc_z / 2, hc_z / 2, 17)
        grid_x = grid_x[5:-5] + target_pos[0] / stride
        grid_y = grid_y[5:-5] + target_pos[1] / stride
        x_offset = grid_x / l
        y_offset = grid_y / l

        grid = np.reshape(np.transpose([np.tile(x_offset, len(y_offset)), np.repeat(y_offset, len(x_offset))]), (len(grid_y), len(grid_x), 2))
        grid = torch.from_numpy(grid).unsqueeze(0).cuda()

        zmap = nn.functional.grid_sample(self.xf.double(), grid).float()
        # cls_kernel = self.rpn.cls.make_kernel(zmap)
        self.MA_kernel = (1 - lambda_u) * self.MA_kernel + lambda_u * zmap
        # self.zf_update = self.zf * lambda_s + self.MA_kernel * (1.0 - lambda_s)
        self.zf = self.zf * lambda_s + self.MA_kernel * (1.0 - lambda_s)

    def track(self, x):
        """
        Please see OceanOnlinePT for pytorch version (more clean)
        """
        if x.size(-1) == 255:
            b1, b2, b3, xf = self.s_backbone_siam255(x, torch.Tensor([]).cuda())
            xf = self.s_neck255(xf)  # b4

            # backbone encode (something is wrong with connect model)
            cls_z0, cls_z1, cls_z2, cls_x0, cls_x1, cls_x2, reg_z0, reg_z1, reg_z2, reg_x0, reg_x1, reg_x2 = self.multiDiCorr255(xf, self.zf)
        
            # correlation
            cls_z = [cls_z0, cls_z1, cls_z2]
            cls_x = [cls_x0, cls_x1, cls_x2]
            reg_z = [reg_z0, reg_z1, reg_z2]
            reg_x = [reg_x0, reg_x1, reg_x2]

            cls_dw, reg_dw = self.connect_model2(cls_z, cls_x, reg_z, reg_x)
            # cls and reg
            bbox_pred, cls_pred = self.boxtower255(cls_dw, reg_dw)
            pred_mask = self.mask_model(b1, b2, b3, xf, cls_dw, x.size()[2:], self.zf_ori, self.template_mask)

        elif x.size(-1) == 287:
            b1, b2, b3, xf = self.s_backbone_siam287(x, torch.Tensor([]))
            xf = self.s_neck287(xf)  # b4

            # backbone encode
            cls_z0, cls_z1, cls_z2, cls_x0, cls_x1, cls_x2, reg_z0, reg_z1, reg_z2, reg_x0, reg_x1, reg_x2 = self.multiDiCorr287(xf, self.zf)

            # correlation
            cls_z = [cls_z0, cls_z1, cls_z2]
            cls_x = [cls_x0, cls_x1, cls_x2]
            reg_z = [reg_z0, reg_z1, reg_z2]
            reg_x = [reg_x0, reg_x1, reg_x2]

            cls_dw, reg_dw = self.connect_model2(cls_z, cls_x, reg_z, reg_x)

            # cls and reg
            bbox_pred, cls_pred = self.boxtower287(cls_dw, reg_dw)
            pred_mask = self.mask_model(b1, b2, b3, xf, cls_dw, x.size()[2:], self.zf_ori, self.template_mask)
        else:
            raise ValueError('unsupported input size')
        self.search_size = xf.size()
        self.xf = xf.detach()
        return cls_pred, bbox_pred.squeeze(0), pred_mask


    def track_pytorch(self, x):
        b1, b2, b3, xf = self.features(x, False)
        xf = self.neck(xf)  # b4

        # backbone encode
        cls_z0, cls_z1, cls_z2, cls_x0, cls_x1, cls_x2, reg_z0, reg_z1, reg_z2, reg_x0, reg_x1, reg_x2 = self.connect_model0(xf, self.zf)

        # correlation
        cls_z = [cls_z0, cls_z1, cls_z2]
        cls_x = [cls_x0, cls_x1, cls_x2]
        reg_z = [reg_z0, reg_z1, reg_z2]
        reg_x = [reg_x0, reg_x1, reg_x2]

        cls_dw, reg_dw = self.connect_model2(cls_z, cls_x, reg_z, reg_x)
        bbox_pred, cls_pred = self.connect_model1(cls_dw, reg_dw)

        # return cls_pred, bbox_pred.squeeze(0), pred_mask
        return None

    # -------------------------------
    # Functions for online
    # -------------------------------
    def extract_for_online(self, x):
        xf = self.s_backbone_online(x, torch.Tensor([1]).cuda())
        return xf










