# -----------------------------------------------------------------------------
# Copyright (c) Microsoft
# Licensed under the MIT License.
# Written by Zhipeng Zhang (zhangzhipeng2017@ia.ac.cn)
# ------------------------------------------------------------------------------
import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from .connect import xcorr_depthwise



class SimiMask(nn.Module):
    def __init__(self, inchannels=256, outchannels=256):
        super(SimiMask, self).__init__()
        self.s_embed = nn.Conv2d(inchannels, outchannels, 1)  # embedding for search feature
        self.t_embed = nn.Conv2d(inchannels, outchannels, 1)  # embeeding for template feature

    def forward(self, xf, zf, zf_mask):
        # xf: [B, C, H, W]
        # zf: [B, C, H, W]
        # zf_mask: [B, H, W]
        # pdb.set_trace()
        xf = self.s_embed(xf)
        zf = self.t_embed(zf)

        B, C, Hx, Wx = xf.size()
        B, C, Hz, Wz = zf.size()

        xf = xf.permute(0, 2, 3, 1).contiguous()  # [B, H, W, C]
        xf = xf.view(B, -1, C)   # [B, H*W, C]
        zf = zf.view(B, C, -1)   # [B, C, H*W]

        att = torch.matmul(xf, zf)  # [HW, HW]
        # att = xf.squeeze().mm(zf.squeeze())
        # att = att.unsqueeze(0)  # [HW, HW]
        att = att / math.sqrt(C)
        att = F.softmax(att, dim=2)  # [HW, HW]
        zf_mask = nn.Upsample(size=(Hz, Wz), mode='bilinear', align_corners=True)(zf_mask.unsqueeze(1))
        # zf_mask = (zf_mask > 0.5).float()
        zf_mask = zf_mask.view(B, -1, 1)

        pred_mask = torch.matmul(att, zf_mask)  # [B, H*W]
        pred_mask = pred_mask.view(B, Hx, Hx).unsqueeze(1)
        return pred_mask

class MultiRefine(nn.Module):
    def __init__(self, att=False, merge_mask=False, addCorr=True, mulOradd='mul', simiLoss=True):
        super(MultiRefine, self).__init__()
        # BACKBONE
        self.b4 = nn.Sequential(nn.Conv2d(256, 64, 3, padding=1), nn.ReLU())
        self.b3 = nn.Sequential(nn.Conv2d(512, 32, 3, padding=1), nn.ReLU())
        self.b2 = nn.Sequential(nn.Conv2d(256, 16, 3, padding=1), nn.ReLU())
        self.b1 = nn.Sequential(nn.Conv2d(64, 4, 3, padding=1), nn.ReLU())
        self.b0 = nn.Sequential(nn.Conv2d(4, 2, 3, padding=1), nn.ReLU())

        # REFINE
        self.rCo = nn.Sequential(nn.Conv2d(256, 64, 3, padding=1), nn.ReLU())
        self.r3 = nn.Sequential(nn.Conv2d(64, 32, 3, padding=1), nn.ReLU())
        self.r2 = nn.Sequential(nn.Conv2d(32, 16, 3, padding=1), nn.ReLU())
        self.r1 = nn.Sequential(nn.Conv2d(16, 4, 3, padding=1), nn.ReLU())
        self.r0 = nn.Sequential(nn.Conv2d(4, 2, 3, padding=1))

        # multi refine layer
        self.m3 = nn.Sequential(nn.Conv2d(32, 2, 3, padding=1))  # stride = 8
        self.m2 = nn.Sequential(nn.Conv2d(16, 2, 3, padding=1))  # stride = 4
        self.m1 = nn.Sequential(nn.Conv2d(4, 2, 3, padding=1))   # stride = 2

        self.att = att
        self.merge_mask = merge_mask
        self.multi_loss = True
        self.simiLoss = simiLoss

        if self.merge_mask:
            self.weight = nn.Parameter(torch.ones(3))
            self.multi_loss = False

        # for mask pretrain
        # self.mask_dw = DepthCorr(256, 256, 2)

        # for sequential
        self.addCorr = addCorr
        self.mulOradd = mulOradd
        self.sequential = SimiMask(256, 64)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_uniform_(m.weight, a=1)

    def forward(self, b1, b2, b3, b4, corr, input_size, zf_ori, template_mask):

        b4_size = b4.size()[2:]
        b3_size = b3.size()[2:]
        b2_size = b2.size()[2:]
        b1_size = b1.size()[2:]
        if input_size is None: input_size = (255, 255)

        # prepare for sequential model
        simiMask = self.sequential(b4, zf_ori, template_mask)  # [B, H, W]
        simiMask = torch.clamp(simiMask, 0, 1)
        if self.mulOradd == 'mul':
            b4 = b4 * simiMask
        elif self.mulOradd == 'add':
            b4 = b4 + simiMask
        else:
            raise ValueError('not implemented mask guidence way')  # maybe cat

        if self.addCorr:
            corr = nn.Upsample(size=b4_size, mode='bilinear', align_corners=True)(corr)
            r4 = self.rCo(corr) + self.b4(b4)
        else:
            r4 = self.b4(b4)

        r4 = nn.Upsample(size=b3_size, mode='bilinear', align_corners=True)(r4)
        r3 = self.r3(r4) + self.b3(b3)

        if self.training:
            m3 = self.m3(r3)   # mask s=8

        if self.att:
            r3 = F.softmax(m3, dim=1)[:, 1, ...].unsqueeze(1) * r3   # attention (mask attention)

        # r3up + b2
        r3 = nn.Upsample(size=b2_size, mode='bilinear', align_corners=True)(r3)
        r2 = self.r2(r3) + self.b2(b2)

        if self.training:
            m2 = self.m2(r2)  # mask s=4

        if self.att:
            r2 = F.softmax(m2, dim=1)[:, 1, ...].unsqueeze(1) * r2 # attention (mask attention)

        # r2up + b1
        r2 = nn.Upsample(size=b1_size, mode='bilinear', align_corners=True)(r2)
        r1 = self.r1(r2) + self.b1(b1)

        if self.training:
            m1 = self.m1(r1)  # mask s=2

        if self.att:
            r1 = F.softmax(m1, dim=1)[:, 1, ...].unsqueeze(1) * r1   # attention (mask attention)

        # r1 up
        r1 = nn.Upsample(size=input_size, mode='bilinear', align_corners=True)(r1)
        mask = self.r0(r1)


        if not self.training:
            return mask

